/* Metageek WiSPY interface 
 * Mike Kershaw/Dragorn <dragorn@kismetwireless.net>
 *
 * This code is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Extra thanks to Ryan Woodings @ Metageek for interface documentation
 */

#include <stdlib.h>
#include "config.h"
#include "wispy_gtk_hw_registry.h"
#include "wispy_gtk.h"
#include "wispy_net.h"

/* Forward defs of button stuff */
void wdr_add_button(wispy_device_registry *wdr, wdr_reg_dev *dev);
void wdr_del_button(wispy_device_registry *wdr, wdr_reg_dev *dev);
void wdr_upd_button(wispy_device_registry *wdr, wdr_reg_dev *dev);

void wdr_init(wispy_device_registry *wdr, GtkWidget *buttonpane) {
	int x = 0;

	wdr->max_dev = WDR_MAX_DEV - 1;
	wdr->cur_dev = 0;

	for (x = 0; x < WDR_MAX_DEV; x++) {
		wdr->devices[x] = NULL;
	}

	wdr->buttonpane = buttonpane;
}

void wdr_free(wispy_device_registry *wdr) {
	int x = 0;

	for (x = 0; x < WDR_MAX_DEV; x++) {
		if (wdr->devices[x] == NULL)
			continue;

		/* Close it */
		wispy_phy_close(wdr->devices[x]->phydev);

		g_list_free(wdr->devices[x]->sweep_cb_l);

		/* remove the polling event */
		gdk_input_remove(wdr->devices[x]->poll_tag);
		free(wdr->devices[x]->poll_rec);

		free(wdr->devices[x]);
	}
}

int wdr_open_add(wispy_device_registry *wdr, wispy_device_rec *devrec,
				 char *errstr) {
	int x;
	wispy_phy *phydev;
	wdr_reg_dev *regdev;

	for (x = 0; x < wdr->max_dev; x++) {
		if (wdr->devices[x] == NULL)
			continue;

		if (wdr->devices[x]->phydev->device_spec->device_id ==
			devrec->device_id) {
			return x;
		}
	}

	if (wdr->cur_dev >= wdr->max_dev) {
		snprintf(errstr, WISPY_ERROR_MAX, "WDR registry full");
		return -1;
	}

	phydev = (wispy_phy *) malloc(WISPY_PHY_SIZE);

	if (wispy_device_init(phydev, devrec) < 0) {
		printf("deug - wdr devinit failed\n");
		snprintf(errstr, WISPY_ERROR_MAX, "Error initializing WiSPY device %s: %s",
				 devrec->name, wispy_get_error(phydev));
		free(phydev);
		return -1;
	}

	if (wispy_phy_open(phydev) < 0) {
		snprintf(errstr, WISPY_ERROR_MAX, "Error opening WiSPY device %s: %s",
				 devrec->name, wispy_get_error(phydev));
		free(phydev);
		return -1;
	}

	/* We always want to calibrate it, imho */
	wispy_phy_setcalibration(phydev, 1);

	regdev = (wdr_reg_dev *) malloc(sizeof(wdr_reg_dev));
	regdev->devbutton = NULL;
	regdev->phydev = phydev;
	regdev->sweep_cb_l = NULL;
	/* The open callback used by the picker/menu needs to claim
	 * a reference to this */
	regdev->refcount = 0;

	/* Queue it for polling */
	regdev->poll_rec = (wdr_poll_rec *) malloc(sizeof(wdr_poll_rec));
	((wdr_poll_rec *) regdev->poll_rec)->wdr = wdr;

	regdev->poll_tag = gdk_input_add(wispy_phy_getpollfd(phydev), 
									 GDK_INPUT_READ, 
									 wdr_poll, regdev->poll_rec);

	((wdr_poll_rec *) regdev->poll_rec)->poll_tag = regdev->poll_tag;

	wdr_add_button(wdr, regdev);

	for (x = 0; x < wdr->max_dev; x++) {
		if (wdr->devices[x] != NULL)
			continue;

		wdr->devices[x] = regdev;
		wdr->cur_dev++;
		((wdr_poll_rec *) regdev->poll_rec)->slot = x;

		return x;
	}

	return x;
}

wispy_phy *wdr_get_phy(wispy_device_registry *wdr, int slot) {
	if (slot < 0 || slot > wdr->max_dev)
		return;

	if (wdr->devices[slot] == NULL)
		return;

	return wdr->devices[slot]->phydev;
}

void wdr_add_ref(wispy_device_registry *wdr, int slot) {
	if (slot < 0 || slot > wdr->max_dev)
		return;

	if (wdr->devices[slot] == NULL)
		return;

	wdr->devices[slot]->refcount++;
}

void wdr_del_ref(wispy_device_registry *wdr, int slot) {
	if (slot < 0 || slot > wdr->max_dev)
		return;

	if (wdr->devices[slot] == NULL)
		return;

	wdr->devices[slot]->refcount--;

	if (wdr->devices[slot]->refcount > 0)
		return;

	wispy_phy_close(wdr->devices[slot]->phydev);

	g_list_free(wdr->devices[slot]->sweep_cb_l);

	/* remove it from the button list */
	wdr_del_button(wdr, wdr->devices[slot]);

	/* remove the polling event */
	gdk_input_remove(wdr->devices[slot]->poll_tag);
	free(wdr->devices[slot]->poll_rec);

	free(wdr->devices[slot]);

	wdr->devices[slot] = NULL;
	wdr->cur_dev--;
}

void wdr_add_sweepcb(wispy_device_registry *wdr, int slot,
					 void (*cb)(int, int, wispy_sample_sweep *, void *),
					 int nagg, void *aux) {
	wdr_reg_sweep_cb *wdrcb;

	if (slot < 0 || slot > wdr->max_dev)
		return;

	if (wdr->devices[slot] == NULL)
		return;

	wdrcb = (wdr_reg_sweep_cb *) malloc(sizeof(wdr_reg_sweep_cb));

	wdrcb->aux = aux;
	wdrcb->cb = cb;

	if (nagg == 0) {
		wdrcb->agg_sweep = NULL;
	} else {
		wdrcb->agg_sweep = wispy_cache_alloc(nagg, 1, 0);
	}
	wdrcb->num_agg = nagg;
	wdrcb->pos_agg = 0;

	wdr->devices[slot]->sweep_cb_l = 
		g_list_append(wdr->devices[slot]->sweep_cb_l, wdrcb);
}

void wdr_del_sweepcb(wispy_device_registry *wdr, int slot,
					 void (*cb)(int, int, wispy_sample_sweep *, void *),
					 void *aux) {
	GList *iter;

	if (slot < 0 || slot > wdr->max_dev)
		return;

	if (wdr->devices[slot] == NULL)
		return;

	iter = wdr->devices[slot]->sweep_cb_l;

	/* Simple iterative search */
	while (iter != NULL) {
		if (((wdr_reg_sweep_cb *) iter->data)->cb == cb &&
			((wdr_reg_sweep_cb *) iter->data)->aux == aux) {
			wdr->devices[slot]->sweep_cb_l =
				g_list_remove(wdr->devices[slot]->sweep_cb_l,
							  iter->data);
			return;
		}

		iter = g_list_next(iter);
	}
}

void wdr_sweep_update(wispy_device_registry *wdr, int slot, 
					  int mode, wispy_sample_sweep *sweep) {
	GList *iter;
	wdr_reg_sweep_cb *scb;

	if (slot < 0 || slot > wdr->max_dev)
		return;

	if (wdr->devices[slot] == NULL)
		return;

	iter = wdr->devices[slot]->sweep_cb_l;

	/* Simple iterative search */
	while (iter != NULL) {
		scb = (wdr_reg_sweep_cb *) iter->data;

		/* Only pass along the peak if we've filled the aggregate or
		 * we've got no aggregation */
		if (scb->agg_sweep == NULL || sweep == NULL) {
			(*(scb->cb))(slot, mode, sweep, scb->aux);
		} else {
			wispy_cache_append(scb->agg_sweep, sweep);
			scb->pos_agg++;

			if (scb->pos_agg == scb->num_agg) {
				(*(scb->cb))(slot, mode, scb->agg_sweep->peak, scb->aux);
				wispy_cache_clear(scb->agg_sweep);
				scb->pos_agg = 0;
			}
		}

		// (*(scb->cb))(slot, mode, sweep, scb->aux);
		iter = g_list_next(iter);
	}
}

void wdr_poll(gpointer data, gint source, GdkInputCondition condition) {
	wdr_poll_rec *wdrpr = (wdr_poll_rec *) data;
	wispy_phy *phydev;
	int r;
	char err[WISPY_ERROR_MAX];
	wispy_sample_sweep *sweep;

	if (wdrpr->slot < 0 || wdrpr->slot > wdrpr->wdr->max_dev)
		return;

	if (wdrpr->wdr->devices[wdrpr->slot] == NULL)
		return;

	phydev = wdrpr->wdr->devices[wdrpr->slot]->phydev;

	do {
		r = wispy_phy_poll(phydev);

		if ((r & WISPY_POLL_ERROR)) {
			snprintf(err, WISPY_ERROR_MAX, "Error polling WiSPY device %s: %s",
					 wispy_phy_getname(phydev), wispy_get_error(phydev));
			wispy_phy_close(phydev);

			/* Push the error, this should result in all the graphs removing
			 * their references to us */
			wdr_sweep_update(wdrpr->wdr, wdrpr->slot, r, NULL);

			/* Pop up the error */
			Wispy_Alert_Dialog(err);

			/* remove us from being pollable, free our tracker ref.  No-one should
			 * reference us anymore since we told them there was an error, so we
			 * can drop out entirely */
			gdk_input_remove(wdrpr->poll_tag);
			free(wdrpr);
			return;
		} else if ((r & WISPY_POLL_CONFIGURED)) {
			wdr_upd_button(wdrpr->wdr, wdrpr->wdr->devices[wdrpr->slot]);
			wdr_sweep_update(wdrpr->wdr, wdrpr->slot, r, NULL);
			return;
		} else if ((r & WISPY_POLL_SWEEPCOMPLETE)) {
			sweep = wispy_phy_getsweep(phydev);

			if (sweep == NULL)
				continue;

			wdr_sweep_update(wdrpr->wdr, wdrpr->slot, r, sweep);

			continue;
		}

	} while ((r & WISPY_POLL_ADDITIONAL));

	/* We do nothing otherwise, no events for partial polls */
}

GList *wdr_populate_menu(wispy_device_registry *wdr, GtkWidget *menu,
					   GCallback cb, void *aux) {
	wdr_menu_rec *r;
	GtkWidget *mi;
	int x;
	GList *cbr = NULL;

	if (wdr->cur_dev <= 0) {
		return NULL;
	}

	for (x = 0; x < wdr->max_dev; x++) {
		if (wdr->devices[x] == NULL)
			continue;

		/* Build the aux */
		r = (wdr_menu_rec *) malloc(sizeof(wdr_menu_rec));
		r->slot = x;
		r->aux = aux;
		r->wdr = wdr;

		/* Add us to the lists */
		cbr = g_list_append(cbr, r);

		/* Increase our refcount - we have to do this so that there is
		 * no race for devices if a graph is being closed */
		wdr_add_ref(wdr, x);

		mi = gtk_menu_item_new_with_label(wispy_phy_getname(wdr->devices[x]->phydev));
		gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
		g_signal_connect_swapped(G_OBJECT(mi), "activate", 
								 G_CALLBACK(cb), r);
		gtk_widget_show(mi);
	}

	mi = gtk_separator_menu_item_new();
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
	gtk_widget_show(mi);

	return cbr;
}

void wdr_free_menu_fe(gpointer *data, gpointer *user) {
	wdr_menu_rec *r = (wdr_menu_rec *) data;

	/* Unref the device */
	wdr_del_ref(r->wdr, r->slot);

	/* Free the data hunk */
	free(r);
}

void wdr_free_menu(wispy_device_registry *wdr, GList *gl) {
	g_list_foreach(gl, (GFunc) wdr_free_menu_fe, NULL);
	g_list_free(gl);
}

void wdr_devpicker_onrowactivated(GtkTreeView *treeview, GtkTreePath *path,
								  GtkTreeViewColumn *column, gpointer *aux) {
	GtkTreeModel *model;
	GtkTreeIter iter;
	wdr_gtk_devpicker_aux *dpaux = (wdr_gtk_devpicker_aux *) aux;
	char err[WISPY_ERROR_MAX];

	model = gtk_tree_view_get_model(treeview);

	if (gtk_tree_model_get_iter(model, &iter, path)) {
		void *phyptr;
		int r;

		gtk_tree_model_get(model, &iter, 1, &phyptr, -1);

		/* Open it and add it to the registry */
		r = wdr_open_add(dpaux->wdr, phyptr, err);

		/* Error alert */
		if (r < 0)
			Wispy_Alert_Dialog(err);

		/* Call their callback saying we've picked a device */
		if (dpaux->pickcb != NULL)
			(*(dpaux->pickcb))(r, dpaux->cbaux);

		gtk_widget_destroy(dpaux->picker_win);
	}
}

void wdr_devpicker_destroy(GtkWidget *widget, gpointer *aux) {
	wdr_gtk_devpicker_aux *dpaux = (wdr_gtk_devpicker_aux *) aux;

	wispy_device_scan_free(&(dpaux->devlist));

	free(aux);
}

void wdr_devpicker_populate(wdr_gtk_devpicker_aux *dpaux) {
	int ndevs, x;
	GtkTreeIter iter;
	GtkListStore *model;
	GtkTreeSelection *selection;

	model = gtk_list_store_new(2, G_TYPE_STRING, G_TYPE_POINTER);
	gtk_tree_view_set_model(GTK_TREE_VIEW(dpaux->treeview), GTK_TREE_MODEL(model));

	/* So that when we replace it in the tree view it gets released */
	g_object_unref(model);

	/*
	if (dpaux->treemodellist != NULL)
		g_free(dpaux->treemodellist);
	*/

	dpaux->treemodellist = model;

	wispy_device_scan_free(&(dpaux->devlist));

	/* Scan for Wispy V1 devices */
	ndevs = wispy_device_scan(&(dpaux->devlist));

	if (ndevs <= 0) {
		gtk_list_store_append(GTK_LIST_STORE(dpaux->treemodellist), &iter);
		gtk_list_store_set(GTK_LIST_STORE(dpaux->treemodellist), &iter, 
						   0, "No WiSPY Devices", 
						   1, NULL,
						   -1);
		gtk_widget_set_sensitive(dpaux->okbutton, 0);
		gtk_widget_set_sensitive(dpaux->scrolled_win, 0);
	} else {
		for (x = 0; x < dpaux->devlist.num_devs; x++) {
			gtk_list_store_append(GTK_LIST_STORE(dpaux->treemodellist), &iter);
			gtk_list_store_set(GTK_LIST_STORE(dpaux->treemodellist), &iter, 
							   0, dpaux->devlist.list[x].name, 
							   1, &(dpaux->devlist.list[x]),
							   -1);

			/* Handle making sure the first item is always selected by
			 * unselecting them all, then selecting one */
			if (x == 0) {
				selection = 
					gtk_tree_view_get_selection(GTK_TREE_VIEW(dpaux->treeview));
				gtk_tree_selection_unselect_all(selection);
				gtk_tree_selection_select_iter(selection, &iter);
			}
		}
		gtk_widget_set_sensitive(dpaux->okbutton, 1);
		gtk_widget_set_sensitive(dpaux->scrolled_win, 1);
	}
}

void wdr_devpicker_button(GtkWidget *widget, gpointer *aux) {
	wdr_gtk_devpicker_aux *dpaux = (wdr_gtk_devpicker_aux *) aux;

	if (widget == dpaux->cancelbutton) {
		gtk_widget_destroy(dpaux->picker_win);
		return;
	}

	if (widget == dpaux->okbutton) {
		GtkTreeSelection *selection;
		GtkTreeModel *model;
		GtkTreeIter iter;
		char err[WISPY_ERROR_MAX];

		selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(dpaux->treeview));
		if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
			void *phyptr;
			int r;
			gtk_tree_model_get(model, &iter, 1, &phyptr, -1);

			/* Open it and add it to the registry */
			r = wdr_open_add(dpaux->wdr, phyptr, err);

			if (r < 0)
				Wispy_Alert_Dialog(err);

			/* Call their callback saying we've picked a device */
			if (dpaux->pickcb != NULL)
				(*(dpaux->pickcb))(r, dpaux->cbaux);
		}

		gtk_widget_destroy(dpaux->picker_win);
		return;
	}

	if (widget == dpaux->rescanbutton) {
		wdr_devpicker_populate(dpaux);
		return;
	}
}

void wdr_devpicker_spawn(wispy_device_registry *wdr,
						 void (*cb)(int, void *),
						 void *aux) {
	GtkWidget *top_window, *scrolled_window;

	GtkWidget *vbox, *hbox;
	GtkWidget *tree_view;

	GtkListStore *model;
	GtkTreeIter iter;
	GtkCellRenderer *cell;
	GtkTreeViewColumn *column;
	int x, ndevs;

	wdr_gtk_devpicker_aux *dpaux = 
		(wdr_gtk_devpicker_aux *) malloc(sizeof(wdr_gtk_devpicker_aux));

	dpaux->wdr = wdr;
	dpaux->pickcb = cb;
	dpaux->cbaux = aux;

	wispy_device_scan_init(&(dpaux->devlist));

	scrolled_window = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),
								   GTK_POLICY_AUTOMATIC,
								   GTK_POLICY_AUTOMATIC);

	tree_view = gtk_tree_view_new();
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolled_window), 
										  tree_view);

	dpaux->treemodellist = NULL;
	dpaux->scrolled_win = scrolled_window;

	cell = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes("USB Devices",
													  cell,
													  "text", 0,
													  NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(tree_view),
								GTK_TREE_VIEW_COLUMN(column));

	g_signal_connect(tree_view, "row-activated", 
					 (GCallback) wdr_devpicker_onrowactivated, dpaux);

	dpaux->treeview = GTK_TREE_VIEW(tree_view);
	dpaux->treecolumn = column;

	top_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(top_window), "WiSPY USB Selection");
	gtk_container_set_border_width(GTK_CONTAINER(top_window), 10);
	gtk_widget_set_size_request(GTK_WIDGET(top_window), 500, 250);

	dpaux->picker_win = top_window;

	g_signal_connect(G_OBJECT(top_window), "destroy",
					 G_CALLBACK(wdr_devpicker_destroy), dpaux);

	vbox = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(top_window), vbox);
	gtk_widget_show(vbox);

	gtk_box_pack_start(GTK_BOX(vbox), scrolled_window, TRUE, TRUE, 2);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(hbox);

	dpaux->okbutton = gtk_toggle_button_new_with_label("Enable");
	gtk_signal_connect(GTK_OBJECT(dpaux->okbutton), "clicked",
					   GTK_SIGNAL_FUNC(wdr_devpicker_button), dpaux);

	dpaux->cancelbutton = gtk_toggle_button_new_with_label("Cancel");
	gtk_signal_connect(GTK_OBJECT(dpaux->cancelbutton), "clicked",
					   GTK_SIGNAL_FUNC(wdr_devpicker_button), dpaux);

	dpaux->rescanbutton = gtk_toggle_button_new_with_label("Re-Scan");
	gtk_signal_connect(GTK_OBJECT(dpaux->rescanbutton), "clicked",
					   GTK_SIGNAL_FUNC(wdr_devpicker_button), dpaux);

	gtk_box_pack_start(GTK_BOX(hbox), dpaux->okbutton, FALSE, FALSE, 2);
	gtk_box_pack_start(GTK_BOX(hbox), dpaux->cancelbutton, FALSE, FALSE, 2);
	gtk_box_pack_start(GTK_BOX(hbox), dpaux->rescanbutton, FALSE, FALSE, 2);

	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 2);

	wdr_devpicker_populate(dpaux);

	gtk_widget_show(tree_view);
	gtk_widget_show(dpaux->okbutton);
	gtk_widget_show(dpaux->cancelbutton);
	gtk_widget_show(dpaux->rescanbutton);

	gtk_widget_show(scrolled_window);
	gtk_widget_show(top_window);
}

void wdr_netpicker_destroy(GtkWidget *widget, gpointer *aux) {
	wdr_gtk_netpicker_aux *npaux = (wdr_gtk_netpicker_aux *) aux;

	printf("debug - netpicker destroy\n");

	free(aux);
}

void wdr_netpicker_button(GtkWidget *widget, gpointer *aux) {
	wdr_gtk_netpicker_aux *npaux = (wdr_gtk_netpicker_aux *) aux;

	if (widget == npaux->cancelbutton) {
		gtk_widget_destroy(npaux->picker_win);
		return;
	}

	if (widget == npaux->okbutton) {
		char err[WISPY_ERROR_MAX];
		char url[1024];
		int r;

		snprintf(url, 1024, "tcp://%s:%s",
				 gtk_entry_get_text(GTK_ENTRY(npaux->hostentry)),
				 gtk_entry_get_text(GTK_ENTRY(npaux->portentry)));

		wispy_device_scan_init(&(npaux->devlist));
		wispy_net_device_scan_url(&(npaux->devlist), url);

		r = wdr_open_add(npaux->wdr, &(npaux->devlist.list[0]), err);

		if (r < 0)
			Wispy_Alert_Dialog(err);

		/* Call their callback saying we've picked a device */
		if (npaux->pickcb != NULL)
			(*(npaux->pickcb))(r, npaux->cbaux);

		wispy_device_scan_free(&(npaux->devlist));

		gtk_widget_destroy(npaux->picker_win);
		return;
	}
}

void wdr_netpicker_spawn(wispy_device_registry *wdr,
						 void (*cb)(int, void *),
						 void *aux) {
	GtkWidget *top_window;

	GtkWidget *vbox, *hbox;

	GtkWidget *label;

	char port[16];

	wdr_gtk_netpicker_aux *npaux = 
		(wdr_gtk_netpicker_aux *) malloc(sizeof(wdr_gtk_netpicker_aux));

	npaux->wdr = wdr;
	npaux->pickcb = cb;
	npaux->cbaux = aux;

	top_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(top_window), "WiSPY Network");
	gtk_container_set_border_width(GTK_CONTAINER(top_window), 10);
	gtk_widget_set_size_request(GTK_WIDGET(top_window), 500, 250);

	npaux->picker_win = top_window;

	g_signal_connect(G_OBJECT(top_window), "destroy",
					 G_CALLBACK(wdr_netpicker_destroy), npaux);

	vbox = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(top_window), vbox);
	gtk_widget_show(vbox);

	label = gtk_label_new("Host:");
	gtk_widget_show(label);
	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 2);

	npaux->hostentry = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(npaux->hostentry), "host");
	gtk_entry_set_editable(GTK_ENTRY(npaux->hostentry), 1);
	gtk_widget_show(npaux->hostentry);
	gtk_box_pack_start(GTK_BOX(vbox), npaux->hostentry, FALSE, TRUE, 2);

	label = gtk_label_new("Port:");
	gtk_widget_show(label);
	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 2);

	npaux->portentry = gtk_entry_new();
	snprintf(port, 16, "%hd", WISPY_NET_DEFAULT_PORT);
	gtk_entry_set_text(GTK_ENTRY(npaux->portentry), port);
	gtk_entry_set_editable(GTK_ENTRY(npaux->portentry), 1);
	gtk_widget_show(npaux->portentry);
	gtk_box_pack_start(GTK_BOX(vbox), npaux->portentry, FALSE, TRUE, 2);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(hbox);

	npaux->okbutton = gtk_toggle_button_new_with_label("Enable");
	gtk_signal_connect(GTK_OBJECT(npaux->okbutton), "clicked",
					   GTK_SIGNAL_FUNC(wdr_netpicker_button), npaux);

	npaux->cancelbutton = gtk_toggle_button_new_with_label("Cancel");
	gtk_signal_connect(GTK_OBJECT(npaux->cancelbutton), "clicked",
					   GTK_SIGNAL_FUNC(wdr_netpicker_button), npaux);

	gtk_box_pack_start(GTK_BOX(hbox), npaux->okbutton, FALSE, FALSE, 2);
	gtk_box_pack_start(GTK_BOX(hbox), npaux->cancelbutton, FALSE, FALSE, 2);

	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 2);

	gtk_widget_show(npaux->okbutton);
	gtk_widget_show(npaux->cancelbutton);

	gtk_widget_show(top_window);
}

static gboolean wdr_devbutton_press(wdr_reg_dev *dev, GdkEvent *event) {
	if (event->type == GDK_BUTTON_PRESS) {
		GdkEventButton *bevent = (GdkEventButton *) event;

		gtk_menu_popup(GTK_MENU(dev->devmenu), NULL, NULL, NULL, NULL, 
					   bevent->button, bevent->time);

		return TRUE;
	}

	return FALSE;
}

void wdr_add_button(wispy_device_registry *wdr, wdr_reg_dev *dev) {
	GtkWidget *mi;

	dev->devbutton = gtk_button_new();
	dev->devmenu = gtk_menu_new();

	gtk_button_set_label(GTK_BUTTON(dev->devbutton),
						 "Configuring...");
	mi = gtk_menu_item_new_with_label("Configuring...");

	gtk_widget_set_sensitive(mi, 0);
	gtk_menu_shell_append(GTK_MENU_SHELL(dev->devmenu), mi);
	gtk_widget_show(mi);

	g_signal_connect_swapped(G_OBJECT(dev->devbutton), "event",
							 G_CALLBACK(wdr_devbutton_press),
							 dev);

	gtk_widget_show(dev->devmenu);
	gtk_widget_show(dev->devbutton);

	gtk_box_pack_start(GTK_BOX(wdr->buttonpane), dev->devbutton,
					   FALSE, FALSE, 1);

	gtk_widget_show(wdr->buttonpane);
}

void wdr_del_button(wispy_device_registry *wdr, wdr_reg_dev *dev) {
	gtk_widget_hide(dev->devbutton);

	gtk_widget_destroy(dev->devbutton);
	dev->devbutton = NULL;

	/* Hide the button bar if we're about to close the last device */
	if (wdr->cur_dev == 1)
		gtk_widget_hide(wdr->buttonpane);
}

void wdr_upd_button(wispy_device_registry *wdr, wdr_reg_dev *dev) {
	GtkWidget *mi;
	char ltxt[256], *rest;
	int res;
	wispy_sample_sweep *ran;

	gtk_widget_destroy(dev->devmenu);

	dev->devmenu = gtk_menu_new();

	if (dev->phydev->state > WISPY_STATE_CONFIGURING) {
		gtk_button_set_label(GTK_BUTTON(dev->devbutton), 
							 wispy_phy_getname(dev->phydev));

		mi = gtk_menu_item_new_with_label(wispy_phy_getname(dev->phydev));
		gtk_widget_set_sensitive(mi, 0);
		gtk_menu_shell_append(GTK_MENU_SHELL(dev->devmenu), mi);
		gtk_widget_show(mi);

		mi = gtk_separator_menu_item_new();
		gtk_menu_shell_append(GTK_MENU_SHELL(dev->devmenu), mi);
		gtk_widget_show(mi);

		ran = &(dev->phydev->device_spec->supported_ranges[0]);

		if (ran->res_hz > 1000 * 1000) {
			res = ran->res_hz / (1000 * 1000);
			rest = "MHz";
		} else if (ran->res_hz > 1000) {
			res = ran->res_hz / 1000;
			rest = "KHz";
		} else {
			res = ran->res_hz;
			rest = "Hz";
		}

		snprintf(ltxt, 256, "Range: %d-%d%s", 
				 ran->start_khz >= 1000 ?  ran->start_khz / 1000 : ran->start_khz,
				 ran->end_khz >= 1000 ? ran->end_khz / 1000 : ran->end_khz,
				 ran->end_khz >= 1000 ? "MHz" : "KHz");
		mi = gtk_menu_item_new_with_label(ltxt);
		gtk_widget_set_sensitive(mi, 0);
		gtk_menu_shell_append(GTK_MENU_SHELL(dev->devmenu), mi);
		gtk_widget_show(mi);

		snprintf(ltxt, 256, "Resolution: %d%s", res, rest);
		mi = gtk_menu_item_new_with_label(ltxt);
		gtk_widget_set_sensitive(mi, 0);
		gtk_menu_shell_append(GTK_MENU_SHELL(dev->devmenu), mi);
		gtk_widget_show(mi);

		snprintf(ltxt, 256, "Samples: %d", ran->num_samples);
		mi = gtk_menu_item_new_with_label(ltxt);
		gtk_widget_set_sensitive(mi, 0);
		gtk_menu_shell_append(GTK_MENU_SHELL(dev->devmenu), mi);
		gtk_widget_show(mi);

	} else {
		gtk_button_set_label(GTK_BUTTON(dev->devbutton),
							 "Configuring...");

		mi = gtk_menu_item_new_with_label("Configuring...");
		gtk_widget_set_sensitive(mi, 0);
		gtk_menu_shell_append(GTK_MENU_SHELL(dev->devmenu), mi);
		gtk_widget_show(mi);
	}
}

