/*
 * This code is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Extra thanks to Ryan Woodings @ Metageek for interface documentation
 */

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <unistd.h>
#include <netdb.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>

#include "wispy_container.h"
#include "wispy_hw_net.h"
#include "wispy_net.h"

#define CLI_BUF_SZ		2048

/* Aux tracking struct for wispy1 characteristics */
typedef struct _wispy_net_aux {
	int sock;

	char hostname[256];
	short int port;

	struct hostent *host;
	struct sockaddr_in servaddr;
	struct sockaddr_in locladdr;

	uint8_t wbuf[CLI_BUF_SZ];
	uint8_t rbuf[CLI_BUF_SZ];

	int write_pos, read_pos, write_fill, read_fill;

	wispy_sample_sweep *sweepbuf;

	wispy_phy *phydev;
} wispy_net_aux;

/* Prototypes */
int wispy_net_open(wispy_phy *);
int wispy_net_close(wispy_phy *);
int wispy_net_poll(wispy_phy *);
int wispy_net_getpollfd(wispy_phy *);
void wispy_net_setcalibration(wispy_phy *, int);
wispy_sample_sweep *wispy_net_getsweep(wispy_phy *);

/* Scan for devices */
int wispy_net_device_scan(wispy_device_list *list) {
	return 0; 
}

int wispy_net_device_scan_url(wispy_device_list *list, char *url) {
	if (list->num_devs == list->max_devs - 1)
		return -1;

	snprintf(list->list[list->num_devs].name, WISPY_PHY_NAME_MAX,
			 "WiSPY net %s", url);
	list->list[list->num_devs].device_id = 0;
	list->list[list->num_devs].init_func = wispy_net_init;
	list->list[list->num_devs].hw_rec = strdup(url);

	list->num_devs++;

	return 1;
}

int wispy_net_init(wispy_phy *phydev, wispy_device_rec *rec) {
	if (rec->hw_rec == NULL)
		return -1;

	return wispy_net_init_url(phydev, (char *) rec->hw_rec);
}

int wispy_net_init_url(wispy_phy *phydev, char *url) {
	wispy_net_aux *auxptr;
	char hoststr[256];
	short int port;
	int ret;

	phydev->device_spec = (wispy_dev_spec *) malloc(sizeof(wispy_dev_spec));

	phydev->device_spec->device_id = 0;
	snprintf(phydev->device_spec->device_name, WISPY_PHY_NAME_MAX,
			 "WiSPY net %s", url);

	phydev->device_spec->device_version = 0;
	phydev->device_spec->num_sweep_ranges = 0;

	auxptr = malloc(sizeof(wispy_net_aux));
	phydev->auxptr = auxptr;

	if ((ret = sscanf(url, "tcp://%256[^:]:%hd", auxptr->hostname, 
					  &(auxptr->port))) == 1) {
		/* Default port */
		printf("debug - using default port\n");
		auxptr->port = WISPY_NET_DEFAULT_PORT;
	} else if (ret != 2) {
		snprintf(phydev->errstr, WISPY_ERROR_MAX,
				 "wispy_net_init_url invalid url, expected tcp://host:port");
		return -1;
	}

	auxptr->host = gethostbyname(auxptr->hostname);

	if (auxptr->host == NULL) {
		snprintf(phydev->errstr, WISPY_ERROR_MAX,
				 "wispy_net_init_url could not resolve host '%s'", 
				 auxptr->hostname);
		return -1;
	}

	auxptr->sock = -1;

	auxptr->read_pos = 0;
	auxptr->read_fill = 0;
	auxptr->write_pos = 0;
	auxptr->write_fill = 0;

	auxptr->sweepbuf = NULL;

	phydev->open_func = &wispy_net_open;
	phydev->close_func = &wispy_net_close;
	phydev->poll_func = &wispy_net_poll;
	phydev->pollfd_func = &wispy_net_getpollfd;
	phydev->setcalib_func = &wispy_net_setcalibration;
	phydev->getsweep_func = &wispy_net_getsweep;

	return 0;
}

int wispy_net_getpollfd(wispy_phy *phydev) {
	wispy_net_aux *auxptr = (wispy_net_aux *) phydev->auxptr;

	return auxptr->sock;
}

int wispy_net_open(wispy_phy *phydev) {
	wispy_net_aux *auxptr = (wispy_net_aux *) phydev->auxptr;
	int save_mode;

	auxptr->servaddr.sin_family = auxptr->host->h_addrtype;
	memcpy((char *) &(auxptr->servaddr.sin_addr.s_addr), 
		   auxptr->host->h_addr_list[0], auxptr->host->h_length);
	auxptr->servaddr.sin_port = htons(auxptr->port);

	if ((auxptr->sock = socket(AF_INET, SOCK_STREAM, 0)) , 0) {
		snprintf(phydev->errstr, WISPY_ERROR_MAX, 
				 "wispy_net_open socket() failed: %s", strerror(errno));
		return -1;
	}

	auxptr->locladdr.sin_family = AF_INET;
	auxptr->locladdr.sin_addr.s_addr = htonl(INADDR_ANY);
	auxptr->locladdr.sin_port = htons(0);

	if (bind(auxptr->sock, (struct sockaddr *) &(auxptr->locladdr),
			 sizeof(auxptr->locladdr)) < 0) {
		snprintf(phydev->errstr, WISPY_ERROR_MAX,
				 "wispy_net_open bind() failed: %s", strerror(errno));
		return -1;
	}

	if (connect(auxptr->sock, (struct sockaddr *) &(auxptr->servaddr),
				sizeof(auxptr->servaddr)) < 0) {
		snprintf(phydev->errstr, WISPY_ERROR_MAX,
				 "wispy_net_open connect() failed: %s", strerror(errno));
		return -1;
	}

	save_mode = fcntl(auxptr->sock, F_GETFL, 0);
	fcntl(auxptr->sock, F_SETFL, save_mode | O_NONBLOCK);

	phydev->state = WISPY_STATE_CONFIGURING;

	return 1;
}

int wispy_net_close(wispy_phy *phydev) {
	wispy_net_aux *aux;
	
	if (phydev == NULL)
		return 0;

	aux = (wispy_net_aux *) phydev->auxptr;

	if (aux == NULL)
		return 0;

	if (aux->sock >= 0) {
		close(aux->sock);
		aux->sock = -1;
	}

	phydev->state = WISPY_STATE_CLOSED;

	return 1;
}

wispy_sample_sweep *wispy_net_getsweep(wispy_phy *phydev) {
	wispy_net_aux *auxptr = (wispy_net_aux *) phydev->auxptr;

	return auxptr->sweepbuf;
}

void wispy_net_setcalibration(wispy_phy *phydev, int in_calib) {
	return;
}

int wispy_net_parse_device(wispy_phy *phydev, wispy_fr_header *header) {
	/* TODO handle more than one device per header */
	wispy_fr_device *dev;
	wispy_fr_sweep *sweep;

	int num_devices;
	int x;
	
	/* header is validated for length at this point */
	num_devices = header->num_blocks;

	if (num_devices < 0)
		return -1;

	if (num_devices > 1)
		fprintf(stderr, "NOTICE - wispy_net driver doesn't handle multiple phydev per "
				"server currently\n");

	dev = (wispy_fr_device *) header->data;

	phydev->device_spec->device_version = dev->device_version;
	phydev->device_spec->device_flags = dev->device_flags;
	phydev->device_spec->device_id = ntohl(dev->device_id);

	phydev->device_spec->supported_ranges = 
		(wispy_sample_sweep *) malloc(WISPY_SWEEP_SIZE(0) *
									  dev->num_sweep_defs);

	phydev->device_spec->num_sweep_ranges = dev->num_sweep_defs;

	for (x = 0; x < dev->num_sweep_defs; x++) {
		sweep = &(dev->sweep_data[wispy_fr_sweep_size(0) * x]);

		phydev->device_spec->supported_ranges[x].start_khz =
			ntohl(sweep->start_khz);
		phydev->device_spec->supported_ranges[x].res_hz =
			ntohl(sweep->res_hz);
		phydev->device_spec->supported_ranges[x].num_samples =
			ntohs(sweep->num_samples);
		phydev->device_spec->supported_ranges[x].end_khz =
			((ntohl(sweep->res_hz) / 1000) * ntohs(sweep->num_samples)) +
			ntohl(sweep->start_khz);
		phydev->device_spec->supported_ranges[x].min_sample =
			ntohs(sweep->min_sample_db) * -1;
		phydev->device_spec->supported_ranges[x].min_sig_report =
			phydev->device_spec->supported_ranges[x].min_sample;
		phydev->device_spec->supported_ranges[x].max_sample =
			ntohs(sweep->max_sample_db) * -1;
	}

	phydev->state = WISPY_STATE_RUNNING;
}

int wispy_net_parse_sweep(wispy_phy *phydev, wispy_fr_header *header) {
	/* TODO - handle multiple sweep records per poll */
	wispy_fr_sweep *sweep;
	wispy_net_aux *auxptr = (wispy_net_aux *) phydev->auxptr;
	int x;

	sweep = (wispy_fr_sweep *) header->data;

	if (ntohl(sweep->device_id) != phydev->device_spec->device_id) {
		return 0;
	}

	if (auxptr->sweepbuf)
		free(auxptr->sweepbuf);

	auxptr->sweepbuf = 
		(wispy_sample_sweep *) malloc(WISPY_SWEEP_SIZE(ntohs(sweep->num_samples)));

	auxptr->sweepbuf->start_khz = ntohl(sweep->start_khz);
	auxptr->sweepbuf->res_hz = ntohl(sweep->res_hz);
	auxptr->sweepbuf->num_samples = ntohs(sweep->num_samples);
	auxptr->sweepbuf->end_khz = 
		((ntohl(sweep->res_hz) / 1000) * ntohs(sweep->num_samples)) +
		ntohl(sweep->start_khz);
	auxptr->sweepbuf->min_sample = ntohs(sweep->min_sample_db) * -1;
	auxptr->sweepbuf->min_sig_report = auxptr->sweepbuf->min_sample;
	auxptr->sweepbuf->max_sample = ntohs(sweep->max_sample_db) * -1;

	for (x = 0; x < auxptr->sweepbuf->num_samples; x++) {
		auxptr->sweepbuf->sample_data[x] =
			sweep->sample_data[x];
	}

	return 1;
}

int wispy_net_poll(wispy_phy *phydev) {
	wispy_net_aux *auxptr = (wispy_net_aux *) phydev->auxptr;
	wispy_fr_header *header;

	int res, additional;

	if (auxptr->sock < 0) {
		return WISPY_POLL_ERROR;
	}

	/* Do we have enough of a frame to even consider it? */
	while (auxptr->read_fill - auxptr->read_pos >= wispy_fr_header_size()) {
		/* is the frame finished? */
		header = (wispy_fr_header *) &(auxptr->rbuf[auxptr->read_pos]);

		if (ntohl(header->sentinel) != WISPY_NET_SENTINEL) {
			auxptr->read_fill = 0;
			auxptr->read_pos = 0;
			return WISPY_POLL_NONE;
		}

		if (ntohs(header->frame_len) > (auxptr->read_fill - auxptr->read_pos)) {
			return WISPY_POLL_NONE;
		}

		if (header->proto_version != WISPY_NET_PROTO_VERSION) {
			printf("debug - wispy_net_poll got a different version %d vs %d, but we'll try anyhow\n", header->proto_version, WISPY_NET_PROTO_VERSION);
		}

		auxptr->read_pos += ntohs(header->frame_len);

		additional = 0;

		if (auxptr->read_pos >= auxptr->read_fill) {
			auxptr->read_pos = 0;
			auxptr->read_fill = 0;
			additional = WISPY_POLL_ADDITIONAL;
		}

		if (header->block_type == WISPY_NET_FRAME_DEVICE) {
			if (wispy_net_parse_device(phydev, header) >= 0)
				return (WISPY_POLL_CONFIGURED | additional);
		} else if (header->block_type == WISPY_NET_FRAME_SWEEP) {
			/* 0 means no sweep */
			if (wispy_net_parse_sweep(phydev, header) > 0)
				return (WISPY_POLL_SWEEPCOMPLETE | additional);
		}
	}

	/* read as much as we can */
	if ((res = read(auxptr->sock, 
					&(auxptr->rbuf[auxptr->read_fill]),
					CLI_BUF_SZ - auxptr->read_fill)) <= 0) {
		if (errno == EAGAIN)
			return WISPY_POLL_NONE;

		snprintf(phydev->errstr, WISPY_ERROR_MAX,
				 "wispy_net_poll read() failed: %s", 
				 strerror(errno));
		return WISPY_POLL_ERROR;
	}
	auxptr->read_fill += res;
		
	return (WISPY_POLL_NONE | WISPY_POLL_ADDITIONAL);
}

