/* Metageek WiSPY network protocol interface 
 * Mike Kershaw/Dragorn <dragorn@kismetwireless.net>
 *
 * Generic WiSPY userspace container for the Metageek WiSPY
 *
 * This code is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Extra thanks to Ryan Woodings @ Metageek for interface documentation
 */

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <unistd.h>
#include <netdb.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <getopt.h>

#include "config.h"
#include "wispy_container.h"
#include "wispy_net.h"

/* Size of the client buffer - a packet should never be this large since
 * it would fragment all over, so this should be fine */
#define CLI_BUF_SZ		2048

typedef struct _wispy_tcpcli {
	int fd;
	uint8_t wbuf[CLI_BUF_SZ];
	uint8_t rbuf[CLI_BUF_SZ];
	int write_pos;
	int read_pos;

	/* Naive model, we assume that we can write faster than we can fill.
	 * Also, we don't mind so much if we lose sweep data, which is the 
	 * only data which ought to really build up over time */
	int write_fill, read_fill;

	struct _wispy_tcpcli *next;
} wispy_tcpcli;

typedef struct _wispy_tcpserv {
	short int port;
	unsigned int maxclients;
	struct sockaddr_in serv_addr;
	int bindfd;
	fd_set master_fds;
	unsigned int maxfd;

	wispy_tcpcli *cli_list;

	wispy_phy *devs;
	int ndev;
} wispy_tcpserv;

int wts_init(wispy_tcpserv *wts) {
	printf("debug - wts init\n");

	wts->port = 0;
	wts->maxclients = 0;
	wts->bindfd = 0;
	FD_ZERO(&(wts->master_fds));
	memset(&(wts->serv_addr), 0, sizeof(wts->serv_addr));
	wts->maxfd = 0;
	wts->cli_list = NULL;
	wts->devs = NULL;
	wts->ndev = 0;
	return 1;
}

int wts_bind(wispy_tcpserv *wts, char *addr, short int port, char *errstr) {
	int sz = 2;

	/* TODO - support binding to an address */

	printf("debug - wts bind\n");

	memset(&(wts->serv_addr), 0, sizeof(wts->serv_addr));

	wts->port = port;

	wts->serv_addr.sin_family = AF_INET;
	wts->serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	wts->serv_addr.sin_port = htons(port);

	if ((wts->bindfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		snprintf(errstr, WISPY_ERROR_MAX, "socket() failed %s", strerror(errno));
		return -1;
	}

	if (setsockopt(wts->bindfd, SOL_SOCKET, SO_REUSEADDR, &sz, sizeof(sz)) < 0) {
		snprintf(errstr, WISPY_ERROR_MAX, "setsockopt() failed %s", strerror(errno));
		return -1;
	}

	if (bind(wts->bindfd, (struct sockaddr *) &(wts->serv_addr), 
			 sizeof(wts->serv_addr)) < 0) {
		snprintf(errstr, WISPY_ERROR_MAX, "bind() failed %s", strerror(errno));
		return -1;
	}

	if (listen(wts->bindfd, 10) < 0) {
		snprintf(errstr, WISPY_ERROR_MAX, "listen() failed %s", strerror(errno));
		return -1;
	}

	FD_SET(wts->bindfd, &(wts->master_fds));

	if (wts->maxfd < wts->bindfd)
		wts->maxfd = wts->bindfd;

	return 1;
}

wispy_tcpcli *wts_accept(wispy_tcpserv *wts, char *errstr) {
	int newfd;
	wispy_tcpcli *tc;
	struct sockaddr_in client_addr;
	socklen_t client_len;
	char inhost[16];
	int save_mode;

	memset(&client_addr, 0, sizeof(struct sockaddr_in));
	client_len = sizeof(struct sockaddr_in);

	if ((newfd = accept(wts->bindfd, (struct sockaddr *) &client_addr,
						&client_len)) < 0) {
		snprintf(errstr, WISPY_ERROR_MAX, "accept() failed %s", strerror(errno));
		return NULL;
	}

	tc = (wispy_tcpcli *) malloc(sizeof(wispy_tcpcli));

	tc->fd = newfd;

	tc->write_pos = 0;
	tc->read_pos = 0;
	tc->write_fill = 0;
	tc->read_fill = 0;

	tc->next = wts->cli_list;
	wts->cli_list = tc;

	save_mode = fcntl(tc->fd, F_GETFL, 0);
	fcntl(tc->fd, F_SETFL, save_mode | O_NONBLOCK);

	FD_SET(newfd, &(wts->master_fds));

	printf("debug - wts accept new client on %d %p\n", newfd, tc);

	return tc;
}

void wts_remove(wispy_tcpserv *wts, wispy_tcpcli *tc) {
	wispy_tcpcli *tci = wts->cli_list;
	wispy_tcpcli *tcb = NULL;

	if (tc->fd >= 0) {
		FD_CLR(tc->fd, &(wts->master_fds));
		close(tc->fd);
	}

	if (tc == tci) {
		wts->cli_list = tci->next;
		free(tc);
		return;
	}

	while (tci != NULL) {
		tcb = tci;

		tci = tci->next;

		if (tci == tc) {
			tcb->next = tci->next;
			free(tci);
		}
	}

}

int wts_fdset(wispy_tcpserv *wts, fd_set *rfd, fd_set *wfd) {
	wispy_tcpcli *tci = wts->cli_list;
	int x;

	FD_SET(wts->bindfd, rfd);

	/* tcp clients */
	while (tci != NULL) {
		FD_SET(tci->fd, rfd);

		if (tci->write_fill > 0) {
			printf("debug - fd %d write fill %d, flag write\n", tci->fd, tci->write_fill);
			FD_SET(tci->fd, wfd);
		}

		if (tci->fd > wts->maxfd)
			wts->maxfd = tci->fd;

		tci = tci->next;
	}

	/* wispy devices */
	for (x = 0; x < wts->ndev; x++) {
		FD_SET(wispy_phy_getpollfd(&(wts->devs[x])), rfd);

		if (wispy_phy_getpollfd(&(wts->devs[x])) > wts->maxfd)
			wts->maxfd = wispy_phy_getpollfd(&(wts->devs[x]));
	}

	return 1;
}

int wts_cli_append(wispy_tcpcli *tci, uint8_t *data, int len, char *errstr) {
	printf("debug - write %d to %d at %d %p\n", len, tci->fd, tci->write_fill, tci);

	if (tci->write_fill + len >= CLI_BUF_SZ) {
		snprintf(errstr, WISPY_ERROR_MAX, "client write buffer %d can't fit %d bytes, "
				 "%d of %d full", tci->fd, len, tci->write_fill, CLI_BUF_SZ);
		return -1;
	}

	memcpy(&(tci->wbuf[tci->write_fill]), data, len);
	tci->write_fill += len;

	printf("debug - client %d appended %d to buffer (%d of %d)\n", tci->fd, len, tci->write_fill, CLI_BUF_SZ);

	return 1;
}

int wts_send_devblock(wispy_tcpserv *wts, wispy_tcpcli *tci, char *errstr) {
	wispy_fr_header *hdr;
	wispy_fr_device *dev;
	wispy_fr_sweep *sweep;

	int devblen = 0, x = 0, r = 0;
	wispy_sample_sweep *ran;

	printf("debug - sending devblock to %d %p\n", tci->fd, tci);

	/* Size of all the sweeps */
	for (x = 0; x < wts->ndev; x++) 
		devblen += wispy_fr_device_size(wts->devs[x].device_spec->num_sweep_ranges);

	/* Big allocation */
	hdr = (wispy_fr_header *) malloc(wispy_fr_header_size() +
									 devblen);

	hdr->sentinel = htonl(WISPY_NET_SENTINEL);
	hdr->frame_len = htons(wispy_fr_header_size() + devblen);
	hdr->proto_version = WISPY_NET_PROTO_VERSION;
	hdr->block_type = WISPY_NET_FRAME_DEVICE;
	hdr->num_blocks = wts->ndev;

	for (x = 0; x < wts->ndev; x++) {
		wispy_phy *phy = &(wts->devs[x]);
		int nsw = (phy->device_spec->num_sweep_ranges);

		dev = (wispy_fr_device *) &(hdr->data[wispy_fr_device_size(nsw) * x]);

		dev->frame_len = htons(wispy_fr_device_size(nsw));
		dev->device_version = phy->device_spec->device_version;
		dev->device_flags = phy->device_spec->device_flags;
		dev->device_id = htonl(phy->device_spec->device_id);
		dev->device_name_len = strlen(wispy_phy_getname(phy));
		snprintf(dev->device_name, 256, "%s", wispy_phy_getname(phy));
		dev->num_sweep_defs = phy->device_spec->num_sweep_ranges;

		for (r = 0; r < dev->num_sweep_defs; r++) {
			ran = &(phy->device_spec->supported_ranges[r]);
			sweep = (wispy_fr_sweep *) &(dev->sweep_data[wispy_fr_sweep_size(0) * r]);

			sweep->frame_len = htons(wispy_fr_sweep_size(0));
			/* noop here */
			sweep->device_id = 0;
			sweep->sweep_type = 0;

			sweep->start_khz = htonl(ran->start_khz);
			sweep->res_hz = htonl(ran->res_hz);
			sweep->num_samples = htons(ran->num_samples);
			sweep->min_sample_db = htons(ran->min_sample * -1);
			sweep->max_sample_db = htons(ran->max_sample * -1);
		}
	}

	if (wts_cli_append(tci, (uint8_t *) hdr,
					   wispy_fr_header_size() + devblen,
					   errstr) < 0)
		return -1;

	free(hdr);
	
	return 1;
}

int wts_send_sweepblock(wispy_tcpserv *wts, 
						wispy_phy *phydev, wispy_sample_sweep *sweep, 
						char *errstr) {
	wispy_fr_header *hdr;
	wispy_fr_sweep *fsweep;
	wispy_tcpcli *tci = NULL;
	int x;

	/* Big allocation */
	hdr = (wispy_fr_header *) malloc(wispy_fr_header_size() +
									 wispy_fr_sweep_size(sweep->num_samples));

	hdr->sentinel = htonl(WISPY_NET_SENTINEL);
	hdr->frame_len = htons(wispy_fr_header_size() + 
						   wispy_fr_sweep_size(sweep->num_samples));
	hdr->proto_version = WISPY_NET_PROTO_VERSION;
	hdr->block_type = WISPY_NET_FRAME_SWEEP;
	hdr->num_blocks = 1;

	fsweep = (wispy_fr_sweep *) hdr->data;

	fsweep->frame_len = htons(wispy_fr_sweep_size(sweep->num_samples));
	fsweep->device_id = htonl(phydev->device_spec->device_id);
	fsweep->sweep_type = WISPY_NET_SWEEPTYPE_CUR;
	fsweep->start_khz = htonl(sweep->start_khz);
	fsweep->res_hz = htonl(sweep->res_hz);
	fsweep->min_sample_db = htons(sweep->min_sample * -1);
	fsweep->max_sample_db = htons(sweep->max_sample * -1);
	fsweep->num_samples = htons(sweep->num_samples);

	for (x = 0; x < sweep->num_samples; x++)
		fsweep->sample_data[x] = sweep->sample_data[x];

	tci = wts->cli_list;
	while (tci != NULL) {
		printf("debug - sending sweepblock to %d %p\n", tci->fd, tci);

		if (wts_cli_append(tci, (uint8_t *) hdr,
						   wispy_fr_header_size() + 
						   wispy_fr_sweep_size(sweep->num_samples),
						   errstr) < 0)
			return -1;

		tci = tci->next;
	}

	free(hdr);
	
	return 1;
}

int wts_poll(wispy_tcpserv *wts, fd_set *rfd, fd_set *wfd, char *errstr) {
	wispy_tcpcli *tci = NULL;
	int x = 0, r = 0;

	tci = wts->cli_list;
	while (tci != NULL) {
		if (FD_ISSET(tci->fd, wfd)) {
			int res;

			printf("debug - should write for %d from %d end %d len %d\n", tci->fd, tci->write_pos, tci->write_fill, tci->write_fill - tci->write_pos);

			if ((res = write(tci->fd, &(tci->wbuf[tci->write_pos]),
							 tci->write_fill - tci->write_pos)) < 0) {
				snprintf(errstr, WISPY_ERROR_MAX, "write() failed on fd %d %s",
						 tci->fd, strerror(errno));
				wts_remove(wts, tci);
				return 0;
			}

			printf("debug - fd %d wrote %d old pos %d newpos %d fill %d\n", tci->fd, res, tci->write_pos, tci->write_pos + res, tci->write_fill); 

			tci->write_pos += res;

			if (tci->write_pos >= tci->write_fill) {
				printf("debug - fd %d flushed buffer, setting to 0\n", tci->fd);
				tci->write_fill = 0;
				tci->write_pos = 0;
			}
		}

		if (FD_ISSET(tci->fd, rfd)) {
			int res;

			printf("debug - reading fd %d fill %d\n", tci->fd, tci->read_fill);

			if ((res = read(tci->fd, 
							&(tci->rbuf[tci->read_fill]),
							CLI_BUF_SZ - tci->read_fill)) <= 0) {
				if (errno == EAGAIN)
					continue;

				snprintf(errstr, WISPY_ERROR_MAX, 
						 "fd %d read error %s\n", tci->fd, strerror(errno));

				wts_remove(wts, tci);
				return 0;
			}
			tci->read_fill += res;
		}

		tci = tci->next;
	}


	if (FD_ISSET(wts->bindfd, rfd)) {
		if ((tci = wts_accept(wts, errstr)) == NULL)
			return -1;

		printf("debug - poll accepted %p\n", tci);

		/* Send them a device block */
		if (wts_send_devblock(wts, tci, errstr) < 0)
			return -1;
	}

	for (x = 0; x < wts->ndev; x++) {
		if (wispy_get_state(&(wts->devs[x])) == WISPY_STATE_ERROR) {
			snprintf(errstr, WISPY_ERROR_MAX, "Wispy phy %d in error state: %s", 
					 x, wispy_get_error(&(wts->devs[x])));
			return -1;
		}

		if (FD_ISSET(wispy_phy_getpollfd(&(wts->devs[x])), rfd) == 0)
			continue;

		do {
			r = wispy_phy_poll(&(wts->devs[x]));

			if ((r & WISPY_POLL_ERROR)) {
				snprintf(errstr, WISPY_ERROR_MAX, "Wispy phy %d poll failed: %s", 
						 x, wispy_get_error(&(wts->devs[x])));
				return -1;
			}

			if ((r & WISPY_POLL_SWEEPCOMPLETE)) {
				if (wts_send_sweepblock(wts, &(wts->devs[x]),
										wispy_phy_getsweep(&(wts->devs[x])),
										errstr) < 0)
					return -1;
			}
		} while ((r & WISPY_POLL_ADDITIONAL));
	}

	return 1;
}

void wts_shutdown(wispy_tcpserv *wts) {
	wispy_tcpcli *tci = wts->cli_list;

	while (tci != NULL) {
		wispy_tcpcli *tcb = tci;
		close(tci->fd);
		tci = tci->next;
		free(tcb);
	}

	close(wts->bindfd);
}

void Usage() {

}

int main(int argc, char *argv[]) {
	wispy_tcpserv wts;
	char errstr[WISPY_ERROR_MAX];
	fd_set sel_r_fds, sel_w_fds;
	struct timeval tm;
	
	wispy_device_list list;
	wispy_phy *devs = NULL;
	int ndev = 0;

	int x = 0, r = 0;

	static struct option long_options[] = {
		{ "port", required_argument, 0, 'p' },
		{ "bindaddr", required_argument, 0, 'a' },
		{ "help", no_argument, 0, 'h' },
		{ 0, 0, 0, 0 }
	};
	int option_index;

	char *bindaddr = NULL;
	int bindport = WISPY_NET_DEFAULT_PORT;

	while (1) {
		int o = getopt_long(argc, argv, "p:a:h",
							long_options, &option_index);

		if (o < 0)
			break;

		if (o == 'h') {
			Usage();
			exit(-1);
		} else if (o == 'a') {
			bindaddr = strdup(optarg);
			continue;
		} else if (o == 'p') {
			if (sscanf(optarg, "%hd", &bindport) != 1) {
				fprintf(stderr, "Expected port number\n");
				Usage();
				exit(-1);
			}
		}
	}

	ndev = wispy_device_scan(&list);

	if (ndev <= 0) {
		printf("No wispy devices found, bailing\n");
		exit(1);
	}

	fprintf(stderr, "Found %d wispy devices...\n", ndev);

	devs = (wispy_phy *) malloc(WISPY_PHY_SIZE * ndev);

	for (x = 0; x < ndev; x++) {
		fprintf(stderr, "Initializing WiSPY device %s id %u\n", 
				list.list[x].name, list.list[x].device_id);

		if (wispy_device_init(&(devs[x]), &(list.list[x])) < 0) {
			fprintf(stderr, "Error initializing WiSPY device %s id %u\n",
					list.list[x].name, list.list[x].device_id);
			fprintf(stderr, "%s\n", wispy_get_error(&(devs[x])));
			exit(1);
		}

		if (wispy_phy_open(&(devs[x])) < 0) {
			fprintf(stderr, "Error opening WiSPY device %s id %u\n",
					list.list[x].name, list.list[x].device_id);
			fprintf(stderr, "%s\n", wispy_get_error(&(devs[x])));
			exit(1);
		}

		wispy_phy_setcalibration(&(devs[x]), 1);
	}
	wispy_device_scan_free(&list);

	wts_init(&wts);

	wts.devs = devs;
	wts.ndev = ndev;

	if (wts_bind(&wts, bindaddr, bindport, errstr) < 0) {
		fprintf(stderr, "TCP bind failed: %s\n", errstr);
		exit(1);
	}

	fprintf(stderr, "TCP server listening on %s port %hd\n",
			bindaddr == NULL ? "(any)" : bindaddr,
			bindport);

	while (1) {
		FD_ZERO(&sel_r_fds);
		FD_ZERO(&sel_w_fds);

		wts_fdset(&wts, &sel_r_fds, &sel_w_fds);

		tm.tv_sec = 0;
		tm.tv_usec = 100000;

		if (select(wts.maxfd + 1, &sel_r_fds, &sel_w_fds, NULL, &tm) < 0) {
			fprintf(stderr, "Select() failed: %s\n", strerror(errno));
			wts_shutdown(&wts);
			exit(1);
		}

		if (wts_poll(&wts, &sel_r_fds, &sel_w_fds, errstr) < 0) {
			fprintf(stderr, "Polling failed: %s\n", errstr);
			wts_shutdown(&wts);
			exit(1);
		}
	}
}

