/* Metageek WiSPY LibUSB interface 
 *
 * LibUSB userspace control for the Metageek WiSPY spectrum analyzer
 * http://www.metageek.net/
 *
 * Basic dumper to dump the raw values from the USB device to stdout.  Meant
 * to be fed to another app via popen (such as the python grapher).  Caller
 * is responsible for converting the RSSI values.
 *
 * Version 1.0
 * Mike Kershaw/Dragorn <dragorn@kismetwireless.net>
 *
 * This code is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Extra thanks to Ryan Woodings @ Metageek for interface documentation
 */

#include <stdio.h>
#include <usb.h>
#include <stdlib.h>
#include <signal.h>
#include <getopt.h>
#include <string.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>

#include "config.h"

#include "wispy_container.h"
#include "wispy_hw_gen1.h"
#include "wispy_hw_net.h"

wispy_phy *devs;
int ndev;

void sighandle(int sig) {
	int x;

	printf("Dying %d from signal %d\n", getpid(), sig);
	for (x = 0; x < ndev; x++) {
		wispy_phy_close(&(devs[x]));
	}
}

void Usage(void) {
	printf("wispy_raw [ options ]\n"
		   " -n / --net  tcp://host:port  Connect to network server instead of\n"
		   "                              local USB devices\n");
	return;
}

int main(int argc, char *argv[]) {
	wispy_device_list list;
	int x = 0, r = 0;
	wispy_sample_sweep *sb;

	static struct option long_options[] = {
		{ "net", required_argument, 0, 'n' },
		{ "help", no_argument, 0, 'h' },
		{ 0, 0, 0, 0 }
	};
	int option_index;

	char *neturl = NULL;

	while (1) {
		int o = getopt_long(argc, argv, "n:h",
							long_options, &option_index);

		if (o < 0)
			break;

		if (o == 'h') {
			Usage();
			return;
		} else if (o == 'n') {
			neturl = strdup(optarg);
			printf("debug - wispy_raw neturl %s\n", neturl);
			continue;
		}
	}

	signal(SIGINT, sighandle);

	if (neturl == NULL) {
		ndev = wispy_device_scan(&list);
		if (ndev <= 0) {
			printf("No wispy devices found, bailing\n");
			exit(1);
		}
	} else {
		wispy_device_scan_init(&list);
		ndev = wispy_net_device_scan_url(&list, neturl);
	}

	printf("Found %d wispy devices...\n", ndev);

	devs = (wispy_phy *) malloc(WISPY_PHY_SIZE * ndev);

	for (x = 0; x < ndev; x++) {
		printf("Initializing WiSPY device %s id %u\n", 
			   list.list[x].name, list.list[x].device_id);

		if (wispy_device_init(&(devs[x]), &(list.list[x])) < 0) {
			printf("Error initializing WiSPY device %s id %u\n",
			   list.list[x].name, list.list[x].device_id);
			printf("%s\n", wispy_get_error(&(devs[x])));
			exit(1);
		}

		if (wispy_phy_open(&(devs[x])) < 0) {
			printf("Error opening WiSPY device %s id %u\n",
			   list.list[x].name, list.list[x].device_id);
			printf("%s\n", wispy_get_error(&(devs[x])));
			exit(1);
		}

		wispy_phy_setcalibration(&(devs[x]), 1);
	}

	wispy_device_scan_free(&list);

	/* Naive poll that doesn't use select() to find pending data */
	while (1) {
		fd_set rfds;
		int maxfd = 0;
		struct timeval tm;

		FD_ZERO(&rfds);

		for (x = 0; x < ndev; x++) {
			FD_SET(wispy_phy_getpollfd(&(devs[x])), &rfds);

			if (wispy_phy_getpollfd(&(devs[x])) > maxfd)
				maxfd = wispy_phy_getpollfd(&(devs[x]));
		}

		tm.tv_sec = 0;
		tm.tv_usec = 10000;

		if (select(maxfd + 1, &rfds, NULL, NULL, &tm) < 0) {
			printf("wispy_raw select() error: %s\n", strerror(errno));
			exit(1);
		}

		for (x = 0; x < ndev; x++) {
			if (wispy_phy_getpollfd(&devs[x]) < 0) {
				printf("Error polling wispy device %s\n",
					   wispy_phy_getname(&(devs[x])));
				printf("%s\n", wispy_get_error(&(devs[x])));
				exit(1);
			}

			if (FD_ISSET(wispy_phy_getpollfd(&(devs[x])), &rfds) == 0)
				continue;

			do {
				r = wispy_phy_poll(&(devs[x]));

				if ((r & WISPY_POLL_CONFIGURED)) {
					printf("Configured wispy %d: %d ranges\n", x, 
						   devs[x].device_spec->num_sweep_ranges);
					for (r = 0; r < devs[x].device_spec->num_sweep_ranges; r++) {
						wispy_sample_sweep *ran = 
							&(devs[x].device_spec->supported_ranges[r]);

						/* Don't do this, kids. */
						printf("    %d: %d%s-%d%s @ %d%s, %d samples\n", r, 
							   ran->start_khz > 1000 ? 
							   ran->start_khz / 1000 : ran->start_khz,
							   ran->start_khz > 1000 ? "MHz" : "KHz",
							   ran->end_khz > 1000 ? ran->end_khz / 1000 : ran->end_khz,
							   ran->end_khz > 1000 ? "MHz" : "KHz",
							   (ran->res_hz / 1000) > 1000 ? (ran->res_hz / 1000) / 1000 : ran->res_hz / 1000,
							   (ran->res_hz / 1000) > 1000 ? "MHz" : "KHz",
							   devs[x].device_spec->supported_ranges[r].num_samples);
					}

					continue;
				} else if ((r & WISPY_POLL_ERROR)) {
					printf("Error polling wispy device %s\n",
						   wispy_phy_getname(&(devs[x])));
					printf("%s\n", wispy_get_error(&(devs[x])));
					exit(1);
				} else if ((r & WISPY_POLL_SWEEPCOMPLETE)) {
					printf("%s sweep: ", wispy_phy_getname(&(devs[x])));
					sb = wispy_phy_getsweep(&(devs[x]));
					if (sb == NULL)
						continue;
					for (r = 0; r < sb->num_samples; r++) {
						printf("%d ", sb->sample_data[r]);
					}
					printf("\n");
				}
			} while ((r & WISPY_POLL_ADDITIONAL));
		}
	}

	return 0;
}	

